#include <stdio.h>
#include <stdlib.h>

#define PARA_SIZE 2500
#define PARA_WORDS 400
#define MIN_LINE_SIZE 55
#define MAX_LINE_SIZE 75
#define OPT_LINE_SIZE 65
#define PARA_LINES (PARA_SIZE/MIN_LINE_SIZE)+1

#define PARAMETER_DEBUG

/* Global variables which are updated during recursion */
int back_prop_level = 0;                  /* The level of recursion */
int line_values[PARA_LINES][PARA_WORDS];  /* Values of previously calculated lines */

/* word_size calculates the length of the word starting at the given
   pointer */
int word_size(word_ptr)
char *word_ptr;  /* A pointer to a "C" style string (terminates with a '\0')
                    which contains the word whose size is to be determined.
                    Size is returned without trailing spaces.                */
  {
  char *curr_pos = word_ptr;  /* Current position in the string */
  int  curr_size = 0;         /* Size of the current word to be returned */

#ifdef PARAMETER_DEBUG
  if (word_ptr == (char *)NULL) return (-1);
#endif

  /* Iterate looking for whitespace or a null character */
  while ((*curr_pos != ' ') && (*curr_pos != '\0'))
    {
    curr_size++;
    curr_pos++;
    }

  return(curr_size);
  }

/* line_size calculates the size of a line */
int line_size(para, words, word_start, word_pos)
char *para;      /* The paragraph in which the line appears */
int  *words,     /* An array of integers containing the positions of word
                   beginnings in the paragraph */
     word_start, /* Starting word for the current line */
     word_pos;   /* Last word in the current line */
  {
  return(words[word_pos] + word_size(&(para[words[word_pos]])) -
         words[word_start]);
  }

/* back_prop calculates the size of the passed line */
int back_prop(para, words, word_start, word_end)
char *para;     /* The paragraph getting margins */
int *words,     /* An array of integers containing the positions of word
                   beginnings in the paragraph */
    word_start, /* Starting word for the current line */
    word_end;   /* Last word in the paragraph */
  {
  int  word_pos = word_start, /* Position of the current end word of the
                                 current line */
       temp_pos,              /* "For" loop temporary variable */
       curr_size=0,           /* The size of the current line */
       next_line_result=0,    /* The back_prop value returned by the next
                                 line */
       best_line_result=0,    /* The best current/next line combination */
       best_current_line=0;   /* The size of the current line which gets
                                 the best line result */

#ifdef PARAMETER_DEBUG
  if (para == (char *)NULL) return (-1);
  if (words == (int *)NULL) return (-2);
  if (word_end < word_start) return (-3);
  if (word_start < 0) return (-4);
  if (word_end < 0) return (-5);
#endif

  /* If the start and end word are the same, then there are no decisions
     to make.  This would be, however, a valid line.                     */
  if (word_start != word_end)
    {
    /* Read up to the minimum size */
    for (temp_pos = word_start;
         ((temp_pos <= word_end) &&
          (words[temp_pos] - words[word_start] < MIN_LINE_SIZE)); temp_pos++)
      {
      word_pos = temp_pos;
      }

    /* Calculate the size of the current line */
    curr_size = line_size(para, words, word_start, word_pos);

    /* Search for the closest to optimum */
    while ((curr_size <= MAX_LINE_SIZE) && (word_pos <= word_end))
      {
      /* Only recurse if this is not the last word of the paragraph */
      if (word_pos != word_end)
        {
        /* check to see if the value has already been calculated */
        if (line_values[back_prop_level+1][word_pos+1] == -1)
          {
          back_prop_level++;
          /* find the best value for the next line */
          next_line_result = back_prop(para, words, word_pos+1, word_end);
          /* track that value so it doesn't have to be calculated again */
          line_values[back_prop_level][word_pos+1] = next_line_result;
          back_prop_level--;
          }
        else
          /* values that have already been calculated can be looked up in the array */
          next_line_result = line_values[back_prop_level+1][word_pos+1];
        }
      else
        /* This means that the next line would be the last line of the paragraph.  Since that line
           can't be optimized, the best solution is to ignore it, which occurs when the result is
           equal to the optimum value */
        next_line_result = OPT_LINE_SIZE;

      /* If the next line ends the paragraph, then it can't approach the
         optimum, so its effects should be ignored. */
      if ((words[word_end] - words[word_start]) <
          (curr_size + next_line_result))
        next_line_result = OPT_LINE_SIZE;

      /* Determine if this is a better line configuration */
      if ((abs(curr_size - OPT_LINE_SIZE) + 
           abs(next_line_result - OPT_LINE_SIZE)) <
          (abs(best_current_line - OPT_LINE_SIZE) +
           abs(best_line_result - OPT_LINE_SIZE)))
        {
        /* Track the best results, so they can be returned */
        best_line_result = next_line_result;
        best_current_line = curr_size;
        }

      /* Add another word if possible */
      word_pos++;
      /* It is only possible to calculate the line size if the starting word is <= to the ending
         word, so only calculate the line size in this case, otherwise there can be strange
         memory problems or negative current size values */
      if (word_pos <= word_end)
        curr_size = line_size(para, words, word_start, word_pos);
      }

    }
  else
    /* Since the start and end are the same, the line result is the size of the word! */
    best_current_line = line_size(para, words, word_start, word_end);

  return(best_current_line);
  }

main()
{
char c,                        /* Current character */
     previous_c,               /* Previous character */
     para[PARA_SIZE+1];        /* Paragraph needing a margin */
int  para_end,                 /* Boolean noting the end of a paragraph */
     para_pos,                 /* Position within the current paragraph */
     para_count,               /* Number of paragraphs processed */
     word_pos,                 /* Current word in the paragraph */
     word_start[PARA_WORDS+1], /* Array of word beginnings in the paragraph */
     temp_pos,                 /* "For" loop variable */
     line_length,              /* Length of the current line */
     line_word,                /* Word starting the current line */
     first_word;               /* First word in the previous line */

para_count = 0;
para_pos = PARA_SIZE;
word_pos = PARA_WORDS;

while (!feof(stdin))
  {
  para_end = 0;
  previous_c = ' ';
  para_count++;

  /* Wipe out the previous paragraph */
  while (para_pos >= 0)
    para[para_pos--] = '\0';
  while (word_pos >= 0)
    word_start[word_pos--] = -1;
  for (temp_pos = 0; temp_pos < PARA_LINES; temp_pos++)
    for (word_pos = 0; word_pos < PARA_WORDS; word_pos++)
      line_values[temp_pos][word_pos] = -1;

  para_pos = 0;
  word_pos = 0;

  while (!feof(stdin) && (!para_end) && 
         (para_pos < PARA_SIZE) && (word_pos < PARA_WORDS))
    {
    c = fgetc(stdin);

    /* Do not reformat any section which has the string "/*" on a line
       beginning.  The section is ended by the end of the file, or by
       "*" followed by "/" on a line.                        */
    if ((previous_c == '/') && (c == '*') && (para_pos == 1))
      {
      if (!feof(stdin))
        previous_c = fgetc(stdin);
      if (!feof(stdin))
        c = fgetc(stdin);

      while (!feof(stdin) && (previous_c != '*') && (c != '/'))
        {
        printf("%c", previous_c);
        previous_c = c;
        c = fgetc(stdin);
        }

      if (!feof(stdin))
        {
        para[0] = '\0';
        para[1] = '\0';
        para_pos = 0;
        word_pos = 0;
        previous_c = ' ';
        c = fgetc(stdin);
        }
      }


    /* Two carriage returns consecutively denote end of paragraph */
    para_end = (previous_c == '\n') && (c == '\n');

    if (c == '\n')
      {
      /* Leave spacing between paragraphs alone */
      if (para_pos == 0)
        {
        printf("\n");
        c = ' ';
        }
      else
        {
        /* Carriage returns are replaced with spaces as line endings will
           be recalculated.                                               */
        if (previous_c != '-')
          {
          /* Quotes can hide end of line punctuation, so ignore quotes when
             considering a second space.                                    */
          if (((previous_c == '"') || (previous_c == '\'')) && (para_pos > 0))
            previous_c = para[para_pos-1];

          para[para_pos++] = ' ';

          /* Perform some checks to see if a double space is needed */
          if ((previous_c == '.') || (previous_c == '?') || 
              (previous_c == '!'))
            para[para_pos++] = ' ';
          }
        }
      }
    else
      {
      /* Every time a new word is encountered it is registered in the 
         word beginnings array.                                       */
      if ((c != ' ') && ((previous_c == '\n') || (previous_c == ' ')))
        word_start[word_pos++] = para_pos;
      para[para_pos++] = c;
      }
    previous_c = c;
    }

  line_word = 0;

  /* Break the paragraph into optimally sized lines and then print the
     new lines out.                                                    */
  while (line_word < word_pos)
    {
    /* NOTE:  Word_pos is actually one longer than the total number
       of words in the word beginning array.  Thus, it is necessary
       to subtract one to pass the last word.                       */
    line_length = back_prop(para, word_start, line_word, word_pos-1);

    /* Print out the current line whose line length has just been
       determined.                                                */
    for (temp_pos=word_start[line_word]; 
         temp_pos < (word_start[line_word] + line_length); temp_pos++)
      printf ("%c", para[temp_pos]);
    printf ("\n");

    /* Check for the end of the paragraph */
    if ((word_start[line_word]+line_length) > (word_start[word_pos-1]))
      {
      /* The last line of the paragraph has been printed */
      line_word = word_pos;
      first_word = line_word;
      }
    else
      first_word = line_word;

    /* Search for the word to start the next paragraph */
    for (temp_pos=first_word; 
         ((temp_pos < word_pos) && 
          (word_start[temp_pos] <= (word_start[first_word] + line_length)));
         temp_pos++)
      {
      line_word = temp_pos;
      }
    line_word++; 
    }

  /* Mark the end of a paragraph with a second carriage return */
  if (!feof(stdin))
    printf("\n");

  if (para_pos == PARA_SIZE)
    printf("Paragraph %d is too large.\n", para_count);
  if (word_pos == PARA_WORDS)
    printf("Paragraph %d has too many words.\n", para_count);
  }


return(0);
}
