//
// Filter.cpp
//

#include "Filter.h"

// Visual Studio only (remove if you are running in other compiler)
// This warning occurs when you have identifiers longer than 255 
// characters, which occurs when you use vector<otherclass>.
#pragma warning(disable:4786)

#ifdef _DEBUG
#define RECALCULATE GenerateString(m_FilterTree, m_wsGeneratedFilter)
#else
#define RECALCULATE { }
#endif

// Errors codes and strings
LdapFilterError LFErrors[] = 
{
    {PARSER_SUCCESS,		    
	L""},
    {LFE_NO_ATTRIBUTE_TYPE,	    
	L"No attribute type specified."},
    {LFE_TYPE_ONLY_SPACES,	    
	L"The attribute type contains only spaces."},
    {LFE_SPACE_MIDDLE_TYPE,	    
	L"Spaces are not allowed in the middle of attribute types."},
    {LFE_NO_ATTRIBUTE_VALUE,	    
	L"No attribute value specified."},
    {LFE_VALUE_ONLY_SPACES,	    
	L"The attribute value contains only spaces."},
    {LFE_ESC_CHAR_WITHOUT_ESCAPABLE,
	L"The character following '\\' isn't an escapable character."},
    {LFE_INVALID_TOKEN,		    
	L"Invalid token."},
    {LFE_BRACKET_MISMATCH,	    
	L"Bracket mismatch."},
    {LFE_NO_VALID_COMPARISON,	    
	L"No comparison exist."},
    {LFE_NOT_SUPPORT_SINGLE_COMP,   
	L"Operator '!' (not) supports only one subfilter."},
    {NULL, NULL}
};


/////////////////////////////////////////////////////////////////
//
// CFilterTree
//
/////////////////////////////////////////////////////////////////

// Constructor
CLdapFilter::CFilterTree::CFilterTree()
{
    m_eOpType = OpTypeNull;
}

// Copy constructor
CLdapFilter::CFilterTree::CFilterTree(const CFilterTree &rhs)
{
    operator=(rhs);
}

// operator=
CLdapFilter::CFilterTree & 
CLdapFilter::CFilterTree::operator=(const CFilterTree &rhs)
{
    m_eOpType = rhs.m_eOpType;
    m_wsItem = rhs.m_wsItem;

    vector<CFilterTree>::const_iterator it;
    it = rhs.m_SubFilters.begin();
    while(it!=rhs.m_SubFilters.end()) {
	// This is a recursive call to operator= !!!
	m_SubFilters.push_back(*it);
	it++;
    }
    
    return *this;
}

// Clear
void CLdapFilter::CFilterTree::Clear()
{
    m_eOpType = OpTypeNull;
    m_SubFilters.erase(m_SubFilters.begin(), m_SubFilters.end());
}


//////////////////////////////////////////////////////
//
// CLdapFilter
//
//////////////////////////////////////////////////////

// Constructor
CLdapFilter::CLdapFilter()
{
    m_fModified = false;
#ifndef LDAPFILTER_SUPPORT_EXCEPTION
    // This member variable exist only if we are not throwing 
    // exception
    m_sLastError = PARSER_SUCCESS;
    m_sErrorPosition = PARSER_SUCCESS;
#endif
}

// Constructor from wstring
CLdapFilter::CLdapFilter(const wstring & rhs)
{
    m_fModified = false;

#ifndef LDAPFILTER_SUPPORT_EXCEPTION
    m_sLastError = PARSER_SUCCESS;
    m_sErrorPosition = PARSER_SUCCESS;
    operator=(rhs);
#else
    try {
	// This operator can throw an exception if the syntax of rhs 
	// is wrong
	operator=(rhs);
    } catch(CFilterException &fe) {
	// if we got this exception, we should clean up the object
	// and re-throw it.
	Clear();
	throw fe;
    }
#endif
    // In Debug mode, this will call GenerateString and update 
    // m_wsFilter
    RECALCULATE;
}

// Constructor from wchar_t *
CLdapFilter::CLdapFilter(const wchar_t * rhs)
{
    m_fModified = false;
#ifndef LDAPFILTER_SUPPORT_EXCEPTION
    m_sLastError = PARSER_SUCCESS;
    m_sErrorPosition = PARSER_SUCCESS;
    operator=(rhs);
#else
    try {
	operator=(rhs);
    } catch(CFilterException &fe) {
	Clear();
	throw fe;
    }
#endif
    RECALCULATE;
}

// Destructor
CLdapFilter::~CLdapFilter()
{
}

// Copy constructor
CLdapFilter::CLdapFilter(const CLdapFilter & rhs)
{
    operator=(rhs);
    RECALCULATE;
}

// operator= 
CLdapFilter & CLdapFilter::operator=(const CLdapFilter & rhs)
{
    if(&rhs==this)
	return *this;

#ifndef LDAPFILTER_SUPPORT_EXCEPTION
    // if we've got errors on rhs, it's a good idea to propagate it.
    m_sLastError = rhs.m_sLastError;
    m_sErrorPosition = rhs.m_sErrorPosition;
#endif

    m_fModified = true; // This object has new content
    m_FilterTree = rhs.m_FilterTree;

    RECALCULATE;
    return *this;
}

// operator= from wstring
CLdapFilter & CLdapFilter::operator=(const wstring & rhs)
{
#ifndef LDAPFILTER_SUPPORT_EXCEPTION
    m_sLastError = PARSER_SUCCESS;
    m_sErrorPosition = PARSER_SUCCESS;
#endif

    // Parse the string, put it into m_FilterTree
    size_t sResult = ParseString(rhs);
    m_fModified = true;
    if(sResult!=PARSER_SUCCESS) {
	// Error
	m_FilterTree.Clear();
	RECALCULATE;
    }
    RECALCULATE;
    return *this;
}

// operator= from wchar_t *
CLdapFilter & CLdapFilter::operator=(const wchar_t * rhs)
{
    if(rhs) {
	return operator=(wstring(rhs));
    } else {
	Clear();
	return *this;
    }
}

// Logical AND
CLdapFilter & CLdapFilter::And(const CLdapFilter & rhs)
{
    // Call the global operator&&(CLdapFilter,CLdapFilter)
    return (*this = operator&&(*this, rhs));
}

CLdapFilter & CLdapFilter::And(const wstring & rhs)
{
    return And(CLdapFilter(rhs));
}

CLdapFilter & CLdapFilter::And(const wchar_t * rhs)
{
    return And(CLdapFilter(rhs));
}

// Logical OR
CLdapFilter & CLdapFilter::Or(const CLdapFilter & rhs)
{
    // Call the global operator||(CLdapFilter,CLdapFilter)
    return (*this = operator||(*this, rhs));
}

CLdapFilter & CLdapFilter::Or(const wstring & rhs)
{
    return Or(CLdapFilter(rhs));
}
CLdapFilter & CLdapFilter::Or(const wchar_t * rhs)
{
    return Or(CLdapFilter(rhs));
}

// Not
CLdapFilter & CLdapFilter::Not()
{
#ifndef LDAPFILTER_SUPPORT_EXCEPTION
    if(m_sLastError!=PARSER_SUCCESS)
	return *this;
#endif

    m_fModified = true;
    if(m_FilterTree.m_eOpType==CFilterTree::OpTypeNot) {
	// If this object was a Not operation, we just put the 
	// subfilter. Example: (!(!(subfilter)) = (subfilter)
	m_FilterTree = *m_FilterTree.m_SubFilters.begin();
    } else if(m_FilterTree.m_eOpType!=CFilterTree::OpTypeNull) {
	// If this was a not, we have to create a new parent
	CFilterTree NewTree;
	NewTree.m_eOpType = CFilterTree::OpTypeNot;
	NewTree.m_SubFilters.push_back(m_FilterTree);
	m_FilterTree = NewTree;
    }

    RECALCULATE;
    return *this;
}

// HasFilter
bool CLdapFilter::HasFilter() const
{
    return m_FilterTree.m_eOpType != CFilterTree::OpTypeNull;
}

// GetString: Return the standard LDAP Filter string
const wstring & CLdapFilter::GetString()
{
    if(m_fModified) {
	// If there was modifications, we have to generate 
	// a new string
	m_wsGeneratedFilter.erase();
	if(m_FilterTree.m_eOpType!=CFilterTree::OpTypeNull) {
	    GenerateString(m_FilterTree, m_wsGeneratedFilter);
	}
    }
    return m_wsGeneratedFilter;
}

// GetCString
// Return the standard LDAP Filter string in a C-style string
const wchar_t * CLdapFilter::GetCString()
{
    return GetString().c_str();
}

// GenerateString
void CLdapFilter::GenerateString(
		    const CLdapFilter::CFilterTree & FilterTree, 
		    wstring & wsFilter)
{
    wsFilter.erase();
    m_fModified = false;

    if(FilterTree.m_eOpType==CFilterTree::OpTypeItem) {
	// This sub-filter is made of only one simple condition, 
	// like (a=b)
	wsFilter = L'(';
	wsFilter += FilterTree.m_wsItem;
	wsFilter += L')';
    } else if(FilterTree.m_eOpType!=CFilterTree::OpTypeNull) {
	// This sub-filter is an operation: And, Or or Not.
	switch(FilterTree.m_eOpType) {
	case CFilterTree::OpTypeNot:
	    wsFilter = L"(!";
	    break;
	case CFilterTree::OpTypeAnd:
	    wsFilter = L"(&";
	    break;
	case CFilterTree::OpTypeOr:
	    wsFilter = L"(|";
	    break;
	}

	wstring wsSubFilter;
	vector<CFilterTree>::const_iterator it;

	it = FilterTree.m_SubFilters.begin();
	while(it != FilterTree.m_SubFilters.end()) {
	    // Recursivelly generate each subfilter under this filter
	    GenerateString(*it, wsSubFilter);
	    wsFilter += wsSubFilter;
	    it++;

	}
	wsFilter += L')';
    }
}

// Clear
void CLdapFilter::Clear()
{
    m_fModified = true;
    m_FilterTree.Clear();
    m_wsGeneratedFilter.erase();
#ifndef LDAPFILTER_SUPPORT_EXCEPTION
    m_sLastError = PARSER_SUCCESS;
    m_sErrorPosition = PARSER_SUCCESS;
#endif

}

// SetLastError
// If LDAPFILTER_SUPPORT_EXCEPTION is defined, we throw an exception,
// else, we set the internal member variables with the error code
// and the position in the string being parsed that the error occured.
size_t CLdapFilter::SetLastError(size_t sPos, size_t sErrorCode)
{
    Clear();
#ifdef LDAPFILTER_SUPPORT_EXCEPTION
    throw CFilterException(sErrorCode, sPos);
#else
    m_sLastError = sErrorCode;
    m_sErrorPosition = sPos;
#endif
    return sPos;
}

#ifndef LDAPFILTER_SUPPORT_EXCEPTION
// The functions ClearLastError, GetLastErrorString, GetLastError 
// and GetLastErrorPos only exist if we are not throwing exceptions.

// ClearLastError
void CLdapFilter::ClearLastError()
{
    m_sLastError = PARSER_SUCCESS;
    m_sErrorPosition = PARSER_SUCCESS;
}

// GetLastErrorString
const wchar_t * CLdapFilter::GetLastErrorString() const
{
    int i = 0;
    while(LFErrors[i].m_sCode) {
	if(LFErrors[i].m_sCode==m_sLastError)
	    return LFErrors[i].m_szError;
	i++;
    }
    return NULL;
}

// GetLastError
size_t CLdapFilter::GetLastError() const
{
    return m_sLastError;
}

// GetLastErrorPos
size_t CLdapFilter::GetLastErrorPos() const
{
    return m_sErrorPosition;
}

#endif


//////////////////////////////////////////////////////
//
// PARSER FUNCTIONS
//
//////////////////////////////////////////////////////


// LdapEscapable
// Parameters:
//   wc1: character to check
//   wc2: second character (optional)
// Return:
//   0: False
//   1 or 2: The first character is, or both are.
// NOTE: Depending on the LDAP server, the Escapable characters can 
//       also be '=', '<', '>', '~', etc.
int CLdapFilter::LdapEscapable(wchar_t wc1, wchar_t wc2)
{
    if(wc1==L'(' || wc1==L')' || wc1==L'*' || wc1==L'\\') {
	return 1;
    }
    return (iswxdigit(wc1) && iswxdigit(wc2))?2:0;
}

// ValidateAttributeType
// Parameters:
//   wsAttr: The attribute type to check. It returns the type
//           without spaces in case of success.
//   sAbsIniPos: The absolute position of wsAttr in the entire 
//               filter being parsed.
// Return:
//   PARSER_SUCCESS: wsAttr is a valid attribute type
//   otherwise: The absolute position that the error occured.
// NOTE: If the LDAPFILTER_SUPPORT_EXCEPTION exist, than SetLastError
//       is going to throw an exception.
// We are only validating if the attribute contains at least one
// character, and do not contain spaces in between characters.
size_t CLdapFilter::ValidateAttributeType(wstring &wsAttr, 
					  size_t sAbsIniPos)
{
    if(wsAttr.size()==0) {
	// Error (No attribute type)
	return SetLastError(sAbsIniPos, LFE_NO_ATTRIBUTE_TYPE);
    }

    // Find the first position not equal space
    size_t sPos = wsAttr.find_first_not_of(L' ');
    if(sPos!=wsAttr.npos) {
	wsAttr.erase(0, sPos); // Remove Leading spaces
    } else {
	// Error (Only spaces)
	return SetLastError(sAbsIniPos, LFE_TYPE_ONLY_SPACES);
    }

    // Find the first space
    size_t sPos2 = wsAttr.find_first_of(L' ');
    if(sPos2==wsAttr.npos)
	return PARSER_SUCCESS;

    size_t sPos3 = wsAttr.find_first_not_of(L' ', sPos2 + 1);
    if(sPos3!=wsAttr.npos) {
	// Error (Spaces in the middle of the attribute type)
	return SetLastError(sAbsIniPos + sPos + sPos2, 
			    LFE_SPACE_MIDDLE_TYPE);
    }

    wsAttr.erase(sPos2); // Remove Trailing spaces

    return PARSER_SUCCESS;

}

// ValidateAttributeValue
// Parameters:
//   wsValue: The attribute value to check. 
//   sAbsIniPos: The absolute position of wsValue in the entire 
//               filter being parsed.
// Return:
//   PARSER_SUCCESS: wsValue is a valid attribute type
//   otherwise: The absolute position that the error occured.
// NOTE: If the LDAPFILTER_SUPPORT_EXCEPTION exist, than SetLastError
//       is going to throw an exception.
// We are only validating if the attribute value is not empty, and
// contains at least one character beside space
size_t CLdapFilter::ValidateAttributeValue(const wstring &wsValue, 
					   size_t sAbsIniPos)
{
    if(wsValue.size()==0) {
	// Error (No value specified)
	return SetLastError(sAbsIniPos, LFE_NO_ATTRIBUTE_VALUE);
    }

    size_t sPos = wsValue.find_first_not_of(L' ');
    if(sPos==wsValue.npos) {
	// Error (Only spaces)
	return SetLastError(sAbsIniPos, LFE_VALUE_ONLY_SPACES);
    }

    return PARSER_SUCCESS;
}

// ParseCondition
// Parameters:
//   wsFilterCond: Simple condition to parse. Returns the condition 
//                 with the attribute type without spaces.
//		   We accept "a=b", "(a=b)" and any extra pair of 
//		   brackets: "(((a=b)))".
//   sAbsIniPos: The absolute position of wsValue in the entire
//               filter being parsed.
// Return:
//   PARSER_SUCCESS: The condition wsFilterCond is valid.
//   otherwise: The absolute position that the error occured.
// This function also throws exception if LDAPFILTER_SUPPORT_EXCEPTION
// is defined.
size_t CLdapFilter::ParseCondition(wstring &wsFilterCond, 
				   size_t sAbsIniPos)
{
    wstring::const_iterator itBegin = wsFilterCond.begin();
    wstring::const_iterator itEnd = wsFilterCond.end();
    wstring::const_iterator it = itBegin;

    // Position of the attribute type in wsFilterCond
    wstring::const_iterator itFirstTypePos = NULL;
    wstring::const_iterator itLastTypePos = NULL;

    // Position of the attribute value in wsFilterCond
    wstring::const_iterator itFirstValuePos = NULL;
    wstring::const_iterator itLastValuePos = NULL;

    int iTotalBracketCount = 0; // Total count of brackets

    wchar_t wcOp = NULL; // Comparisons being performed, 
			 // can be "=", "<", ">" or "~"
    wchar_t wcLast = NULL; // Previous character

    int iSkip = 0;

    while(*it) {
	switch(*it) {
	case L' ':
	    it++;
	    continue;
	    break;
	case L'\\':	// Escape character for LDAP
	    if(wcLast==L')') {
		// Error (Invalid preceding character) - Ex: "(a=b)\"
		return SetLastError(sAbsIniPos + (it - itBegin), 
				    LFE_INVALID_TOKEN);
	    }
	    if(it < itEnd - 2)
		iSkip = LdapEscapable(*(it+1), *(it+2));
	    else
		iSkip = LdapEscapable(*(it+1));
	    if(iSkip==0) {
		// Error (Escape character no followed by a escapable
		//        char) - Ex: "(a=\2g)"
		return SetLastError(sAbsIniPos + (it - itBegin), 
				    LFE_ESC_CHAR_WITHOUT_ESCAPABLE);
	    }
	    if(itFirstTypePos==NULL)
		itFirstTypePos = it;
	    break;
	case L'~': // Comparisons
	case L'>':
	case L'<':
	    if(wcLast==L'(' || wcLast==L')') {
		// Error (Invalid next character) - Ex: "(<=b)"
		return SetLastError(sAbsIniPos + (it - itBegin), 
				    LFE_INVALID_TOKEN);
	    }
	    if(*(it+1)==L'=' && wcOp==NULL) {
		wcOp = *it;
		if(itFirstTypePos)
		    itLastTypePos = it;
		it++; // Next char is a "=", skip it.
		itFirstValuePos = it + 1;
	    }
	    break;
	case L'=':
	    if(wcLast==L'(' || wcLast==L')') {
		// Error (Invalid next character) - Ex: "(=b)"
		return SetLastError(sAbsIniPos + (it - itBegin), 
				    LFE_INVALID_TOKEN);
	    }
	    if(wcOp==NULL) {
		wcOp = L'=';
		itFirstValuePos = it + 1;
		if(itFirstTypePos)
		    itLastTypePos = it;
	    }
	    break;
	case L'(':
	    if(wcLast!=L'(' && wcLast!=NULL ) {
		// Error (Invalid preceding character) Ex: "(a(a=b))"
		return SetLastError(sAbsIniPos + (it - itBegin), 
				    LFE_INVALID_TOKEN);
	    }
	    iTotalBracketCount++;
	    break;
	case L')':
	    if(wcLast==L'(' || wcLast==NULL) {
		// Error (Invalid preceding character) - Ex: "(a=)"
		return SetLastError(sAbsIniPos + (it - itBegin), 
				    LFE_INVALID_TOKEN);
	    }
	    if(itLastValuePos==NULL || itFirstValuePos)
		itLastValuePos = it;
	    iTotalBracketCount--;
	    if(iTotalBracketCount < 0) {
		// Error (Bracket mismatch) - Ex: "(a=b))"
		return SetLastError(sAbsIniPos + (it - itBegin), 
				    LFE_BRACKET_MISMATCH);
	    }
	    break;
	default:
	    if(wcLast==L')') {
		// Error (Invalid preceding character) - Ex: "(a=b)c"
		return SetLastError(sAbsIniPos + (it - itBegin), 
				    LFE_INVALID_TOKEN);
	    }
	    if(itFirstTypePos==NULL)
		itFirstTypePos = it;
	    break;

	}
	wcLast = *it;
	while(iSkip) {
	    iSkip--;
	    it++;
	}
	it++;
    }

    if(iTotalBracketCount!=0) {
	// Error (Bracket mismatch) - Ex: "((a=b)"
	return SetLastError(sAbsIniPos + (it - itBegin),
			    LFE_BRACKET_MISMATCH);
    }

    if(!wcOp || !itFirstTypePos || !itLastTypePos) {
	// Error (No valid comparison) - Ex: "(a)"
	return SetLastError(sAbsIniPos, LFE_NO_VALID_COMPARISON);
    }

    size_t sResult;
    
    // Get the attribute type and validate it
    wstring wsType(itFirstTypePos, itLastTypePos);
    sResult = ValidateAttributeType(wsType, 
			    sAbsIniPos + (itFirstTypePos - itBegin));
    if(sResult!=PARSER_SUCCESS)
	return sResult;

    // Get the attribute value and validate it
    if(itLastValuePos==NULL)
	itLastValuePos = wsFilterCond.end();
    wstring wsValue(itFirstValuePos, itLastValuePos);

    sResult = ValidateAttributeValue(wsValue, 
			    sAbsIniPos + (itFirstValuePos - itBegin));
    if(sResult!=PARSER_SUCCESS)
	return sResult;

    wsFilterCond = wsType;
    if(wcOp!=L'=')
	wsFilterCond += wcOp;
    wsFilterCond += L'=';
    wsFilterCond += wsValue;

    return PARSER_SUCCESS;
}

// FullParser
// Parameters:
//   wsFilter: Filter to parse.
//   FilterTree: Structure that receives the result
//   sAbsIniPos: The absolute position of wsValue in the entire
//               filter being parsed.
// Return:
//   PARSER_SUCCESS: The condition wsFilterCond is valid.
//   otherwise: The absolute position that the error occured.
// This function also throws exception if LDAPFILTER_SUPPORT_EXCEPTION
// is defined.
size_t CLdapFilter::FullParser(const wstring &wsFilter, 
			       CLdapFilter::CFilterTree & FilterTree, 
			       size_t sAbsIniPos)
{
    FilterTree.Clear();

    wstring::const_iterator itBegin = wsFilter.begin();
    wstring::const_iterator itEnd = wsFilter.end();

    wstring::const_iterator it = itBegin;

    wstring::const_iterator itPar = NULL;
    wstring::const_iterator itFirstOperationPar = NULL;

    bool fOperation = false; // If we found an operation (&, | or !)
    bool fFinish = false;    // If we found the last close bracket
    
    int iSkip = 0;	     // Used to skip characters due to the 
			     // Escape char '\'

    int iTotalBracketCount = 0;  // Total brackets count
    int iOpBracketCount = 0;       // Bracket count inside a subfilter

    wchar_t wcLast = NULL;
    bool fLastWasChar = false;
    
    while(*it) {
	
	switch(*it) {
	case L' ':  // Ignore spaces
	    it++;
	    continue;
	    break;
	case L'\\':// If is the escape-leading character...
	    if(wcLast!=L'(' && !fLastWasChar && wcLast!=NULL) {
		// Error (Invalid preceding character) - Ex: "(\a=b))"
		return SetLastError(sAbsIniPos + (it - itBegin), 
				    LFE_INVALID_TOKEN);
	    }
	    if(it < itEnd - 2)
		iSkip = LdapEscapable(*(it+1), *(it+2));
	    else
		iSkip = LdapEscapable(*(it+1));
	    if(iSkip==0) {
		// Error (Escape character no followed by a escapable
		//        char) - Ex: "(a=\2g)"
		return SetLastError(sAbsIniPos + (it - itBegin), 
				    LFE_ESC_CHAR_WITHOUT_ESCAPABLE);
	    }
	    fLastWasChar = true;
	    break;
	case L'(':
	    if(fLastWasChar) {
		// Error (Invalid preceding character) - Ex: "a(a=b)"
		return SetLastError(sAbsIniPos + (it - itBegin), 
				    LFE_INVALID_TOKEN);
	    }
	    if(fOperation) { // If we are on mode operation
		if(iOpBracketCount==0) { // And this is the first bracket 
				   // in a subfilter
		    if(itFirstOperationPar != NULL &&
			FilterTree.m_eOpType==CFilterTree::OpTypeNot) {
			// Error ('Not' supports only a single 
			//        subfilter) - Ex: "(!(a=b)(c=d))"
			return SetLastError(sAbsIniPos + (it-itBegin),
					LFE_NOT_SUPPORT_SINGLE_COMP);
		    }
		    itFirstOperationPar = it;
		} 
		iOpBracketCount++;
	    }
	    if(itPar==NULL)
		itPar = it;
	    iTotalBracketCount++;
	    fLastWasChar = false;
	    break;
	case L')':
	    if(!fLastWasChar && wcLast!=L')') {
		// Error (Invalid preceding character)
		return SetLastError(sAbsIniPos + (it - itBegin),
				    LFE_INVALID_TOKEN);
	    }
	    iTotalBracketCount--;
	    if(iTotalBracketCount==0) {
		fFinish = true;
	    } else if(iTotalBracketCount < 0) {
		// Error (Brackets mismatch)
		return SetLastError(sAbsIniPos + (it - itBegin),
				    LFE_BRACKET_MISMATCH);
	    }
	    
	    if(fOperation) {
		if(!fFinish) {
		    iOpBracketCount--;
		    if(iOpBracketCount==0) {
			// This is our sub-filter to parse
			wstring wsNewExpr(itFirstOperationPar, it+1);
			size_t sResult;
			CFilterTree NewFilterTree;
			// Do a recursive call, and parse this 
			// subfilter
			sResult = FullParser(wsNewExpr, NewFilterTree,
			    sAbsIniPos + (it - itFirstOperationPar));
			if(sResult!=PARSER_SUCCESS)
			    return sResult;
			// If the subfilter has the same comparison
			// operation as FilterTree, do a merge of both
			if(NewFilterTree.m_eOpType==
					FilterTree.m_eOpType) {
			    if(NewFilterTree.m_eOpType==
					CFilterTree::OpTypeNot) {
				// Specially for '!', if both are 
				// equals, means that one
				// neutralize the other.
				FilterTree = 
				    NewFilterTree.m_SubFilters[0];
			    } else {
				FilterTree.m_SubFilters.insert(
				    FilterTree.m_SubFilters.end(),
				    NewFilterTree.m_SubFilters.begin(), 
				    NewFilterTree.m_SubFilters.end());
			    }
			} else {
			    FilterTree.m_SubFilters.push_back(
						NewFilterTree);
			}
		    }
		}
	    } else {
		if(fFinish) {
		    // If we got here is because we didn't find any 
		    // operation, like '&', '|' or '!'
		    // So this is (should be) a simple condition
		    wstring wsSingleCond = wsFilter;
		    
		    size_t sResult = ParseCondition(wsSingleCond, 
						    sAbsIniPos);
		    if(sResult!=PARSER_SUCCESS)
			return sResult;
		    FilterTree.m_eOpType = CFilterTree::OpTypeItem;
		    FilterTree.m_wsItem = wsSingleCond;
		} 
	    }
	    fLastWasChar = false;
	    break;
	case L'&':
	    fLastWasChar = true;
	    if(wcLast==L'(') {
		wstring::const_iterator itTemp = it + 1;
		while(*itTemp==L' ')
		    itTemp++;
		if(*itTemp==L'(') {
		    if(!fOperation) {
			FilterTree.m_eOpType = CFilterTree::OpTypeAnd;
			fOperation = true;
		    }
		    fLastWasChar = false;
		} 
	    }
	    break;
	case L'|':
	    fLastWasChar = true;
	    if(wcLast==L'(') {
		wstring::const_iterator itTemp = it + 1;
		while(*itTemp==L' ')
		    itTemp++;
		if(*itTemp==L'(') {
		    if(!fOperation) {
			FilterTree.m_eOpType = CFilterTree::OpTypeOr;
			fOperation = true;
		    }
		    fLastWasChar = false;
		} 
	    }
	    break;
	case L'!':
	    fLastWasChar = true;
	    if(wcLast==L'(') {
		wstring::const_iterator itTemp = it + 1;
		while(*itTemp==L' ')
		    itTemp++;
		if(*itTemp==L'(') {
		    if(!fOperation) {
			FilterTree.m_eOpType = CFilterTree::OpTypeNot;
			fOperation = true;
		    }
		    fLastWasChar = false;
		} 
	    }
	    break;
	default:
	    if(wcLast!=L'(' && !fLastWasChar && wcLast!=NULL) {
		// Error (Invalid preceding character) - Ex: "(a=b)a"
		return SetLastError(sAbsIniPos + (it - itBegin), 
				    LFE_INVALID_TOKEN);
	    }
	    fLastWasChar = true;
	    break;
	}

	wcLast = *it;
	while(iSkip) {
	    iSkip--;
	    it++;
	}
	it++;
    }
    
    if(iTotalBracketCount!=0) {
	// Error (Bracket mismatch) - Ex: "((a=b)"
	return SetLastError(sAbsIniPos + (it - itBegin),
			    LFE_BRACKET_MISMATCH);
    }

    if(itPar==NULL) {
	// Could be a simple condition, like "a=b"
	wstring wsTemp(wsFilter);
	size_t sResult = ParseCondition(wsTemp, sAbsIniPos);
	if(sResult!=PARSER_SUCCESS)
	    return sResult;
	FilterTree.m_eOpType = CFilterTree::OpTypeItem;
	FilterTree.m_wsItem = wsTemp;
    }

    return PARSER_SUCCESS;
	
}

// ParseString
size_t CLdapFilter::ParseString(const wstring & wsFilter)
{
    if(wsFilter.size()==0) {
	Clear();
	return PARSER_SUCCESS;
    }
    return FullParser(wsFilter, m_FilterTree, 0);
}

///////////////////////////////////////////////////////////////
//
// Friends
//
///////////////////////////////////////////////////////////////

// operator&&
CLdapFilter operator&&(const CLdapFilter lhs, const CLdapFilter & rhs)
{
    CLdapFilter Result;

#ifndef LDAPFILTER_SUPPORT_EXCEPTION
    // if we are not throwing Exceptions, we should propagate erros
    if(lhs.m_sLastError!=PARSER_SUCCESS) {
	Result.m_sLastError = lhs.m_sLastError;
	Result.m_sErrorPosition = 0;
	return Result;
    }
    if(rhs.m_sLastError!=PARSER_SUCCESS) {
	Result.m_sLastError = rhs.m_sLastError;
	Result.m_sErrorPosition = 0;
	return Result;
    }
#endif

    bool fLhsIsNull = false;
    bool fRhsIsNull = false;

    if(lhs.m_FilterTree.m_eOpType ==
		CLdapFilter::CFilterTree::OpTypeNull)
	fLhsIsNull = true; // No filter present in lhs
    if(rhs.m_FilterTree.m_eOpType ==
		CLdapFilter::CFilterTree::OpTypeNull)
	fRhsIsNull = true; // No filter present in rhs

    if(fLhsIsNull && fRhsIsNull) {
	// If both are empty, we return empty
	return Result;
    }

    if(fRhsIsNull) {
	// If rhs is empty, we return lhs
	Result = lhs;
	return Result;
    }
    if(fLhsIsNull) {
	// if lhs is empty, we return rhs
	Result = rhs;
	return Result;
    }

    // The new filter is an AND
    Result.m_FilterTree.m_eOpType = 
		    CLdapFilter::CFilterTree::OpTypeAnd;

    if(lhs.m_FilterTree.m_eOpType ==
		    CLdapFilter::CFilterTree::OpTypeAnd) {
	// If lhs is also an AND, we merge subfilters
	vector<CLdapFilter::CFilterTree>::const_iterator it;
	it = lhs.m_FilterTree.m_SubFilters.begin();
	while(it!=lhs.m_FilterTree.m_SubFilters.end()) {
	    Result.m_FilterTree.m_SubFilters.push_back(*it);
	    it++;
	}
    } else {
	// Need to add the entire lhs to the Subfilters vector
	Result.m_FilterTree.m_SubFilters.push_back(lhs.m_FilterTree);
    }
		
    if(rhs.m_FilterTree.m_eOpType ==
		    CLdapFilter::CFilterTree::OpTypeAnd) {
	// If rhs is also an AND, we merge subfilters
	vector<CLdapFilter::CFilterTree>::const_iterator it;
	it = rhs.m_FilterTree.m_SubFilters.begin();
	while(it!=rhs.m_FilterTree.m_SubFilters.end()) {
	    Result.m_FilterTree.m_SubFilters.push_back(*it);
	    it++;
	}
    } else {
	// Need to add the entire rhs to the Subfilters vector
	Result.m_FilterTree.m_SubFilters.push_back(rhs.m_FilterTree);
    }

    return Result;
}

// operator||
CLdapFilter operator||(const CLdapFilter lhs, 
		       const CLdapFilter & rhs)
{
    CLdapFilter Result;

#ifndef LDAPFILTER_SUPPORT_EXCEPTION
    // if we are not throwing Exceptions, we should propagate erros
    if(lhs.m_sLastError!=PARSER_SUCCESS) {
	Result.m_sLastError = lhs.m_sLastError;
	Result.m_sErrorPosition = 0;
	return Result;
    }
    if(rhs.m_sLastError!=PARSER_SUCCESS) {
	Result.m_sLastError = rhs.m_sLastError;
	Result.m_sErrorPosition = 0;
	return Result;
    }
#endif

    bool fLhsIsNull = false;
    bool fRhsIsNull = false;

    if(lhs.m_FilterTree.m_eOpType ==
		    CLdapFilter::CFilterTree::OpTypeNull)
	fLhsIsNull = true; // No filter present in lhs
    if(rhs.m_FilterTree.m_eOpType ==
		    CLdapFilter::CFilterTree::OpTypeNull)
	fRhsIsNull = true; // No filter present in rhs

    if(fLhsIsNull && fRhsIsNull) {
	// If both are empty, we return empty
	return Result;
    }

    if(fRhsIsNull) {
	// If rhs is empty, we return lhs
	Result = lhs;
	return Result;
    }
    if(fLhsIsNull) {
	// If lhs is empty, we return rhs
	Result = rhs;
	return Result;
    }

    // The new filter is an OR
    Result.m_FilterTree.m_eOpType = 
		    CLdapFilter::CFilterTree::OpTypeAnd;

    if(lhs.m_FilterTree.m_eOpType ==
		    CLdapFilter::CFilterTree::OpTypeAnd) {
	// If lhs is also an OR, we merge subfilters
	vector<CLdapFilter::CFilterTree>::const_iterator it;
	it = lhs.m_FilterTree.m_SubFilters.begin();
	while(it!=lhs.m_FilterTree.m_SubFilters.end()) {
	    Result.m_FilterTree.m_SubFilters.push_back(*it);
	    it++;
	}
    } else {
	// Need to add the entire lhs to the Subfilters vector
	Result.m_FilterTree.m_SubFilters.push_back(lhs.m_FilterTree);
    }
		
    if(rhs.m_FilterTree.m_eOpType ==
		    CLdapFilter::CFilterTree::OpTypeAnd) {
	// If rhs is also an OR, we merge subfilters
	vector<CLdapFilter::CFilterTree>::const_iterator it;
	it = rhs.m_FilterTree.m_SubFilters.begin();
	while(it!=rhs.m_FilterTree.m_SubFilters.end()) {
	    Result.m_FilterTree.m_SubFilters.push_back(*it);
	    it++;
	}
    } else {
	// Need to add the entire rhs to the Subfilters vector
	Result.m_FilterTree.m_SubFilters.push_back(rhs.m_FilterTree);
    }

    return Result;
}


// operator!
CLdapFilter operator!(const CLdapFilter &rhs)
{
    CLdapFilter Result;

#ifndef LDAPFILTER_SUPPORT_EXCEPTION
    // if we are not throwing Exceptions, we should propagate erros
    if(rhs.m_sLastError!=PARSER_SUCCESS) {
	Result.m_sLastError = rhs.m_sLastError;
	Result.m_sErrorPosition = 0;
	return Result;
    }
#endif

    if(rhs.m_FilterTree.m_eOpType ==
		    CLdapFilter::CFilterTree::OpTypeNot) {
	// If this object was a 'Not' operation, we just pick the
	// subfilter. Example: (!(!(subfilter)) = (subfilter)
	Result.m_FilterTree = *rhs.m_FilterTree.m_SubFilters.begin();
    } else if(rhs.m_FilterTree.m_eOpType != 
			CLdapFilter::CFilterTree::OpTypeNull) {
	// If this was a not, we have to create a new parent
	CLdapFilter::CFilterTree NewTree;
	NewTree.m_eOpType = CLdapFilter::CFilterTree::OpTypeNot;
	NewTree.m_SubFilters.push_back(rhs.m_FilterTree);
	Result.m_FilterTree = NewTree;
	
    }

    return Result;
}
