//
// Filter3.h
//
// To support Excetpion, compile with 
//  #define LDAPFILTER_SUPPORT_EXCEPTION
//

#ifndef __FILTER3_H__
#define __FILTER3_H__

// Comment the following line if you don't want the CLdapFilter
// to throw exceptions. Then, you'll be able to use the functions 
// GetLastError* to get the error code
//#define LDAPFILTER_SUPPORT_EXCEPTION

// Visual Studio only (remove if you are running in other compiler)
#pragma warning(disable:4786)

#include <string>
#include <vector>
#include "lferror.h"

// Make the wstring and vector available in the global namespace
using namespace std;

#ifdef LDAPFILTER_SUPPORT_EXCEPTION
#define LFE_THROW throw();
#else
#define LFE_THROW
#endif

// Use this to construct primitive filters, like F("sn=Wood")
#define F(x) CLdapFilter(L## x)


class CLdapFilter {

    // Friends
    friend CLdapFilter operator&&(const CLdapFilter lhs, 
				  const CLdapFilter & rhs);
    friend CLdapFilter operator||(const CLdapFilter lhs, 
				  const CLdapFilter & rhs);
    friend CLdapFilter operator!(const CLdapFilter &rhs);

    // Struct CFilterTree
    struct CFilterTree {
	enum eOpType {
	    OpTypeNull, // No operation
	    OpTypeItem, // Single Item
	    OpTypeAnd,  // Logical AND
	    OpTypeOr,   // Logical OR
	    OpTypeNot   // Logical NOT
	};

	CFilterTree();
	CFilterTree(const CFilterTree &rhs);
	CFilterTree & operator=(const CFilterTree &rhs);

	void Clear();

	eOpType m_eOpType;
	// if m_eOpType is OpTypeItem we use the member m_wsItem, 
	// if not we use m_SubFilters
	vector<CFilterTree> m_SubFilters;
	wstring m_wsItem;
    };

public:
    // Default constructor and destructor
    CLdapFilter();
    ~CLdapFilter();

    // Others constructors
    CLdapFilter(const CLdapFilter & rhs);
    CLdapFilter(const wstring & rhs) LFE_THROW;
    CLdapFilter(const wchar_t * rhs) LFE_THROW;

    // Assignment operator
    CLdapFilter & operator=(const CLdapFilter & rhs);
    CLdapFilter & operator=(const wstring & rhs) LFE_THROW;
    CLdapFilter & operator=(const wchar_t * rhs) LFE_THROW;

    // Logical AND
    CLdapFilter & And(const CLdapFilter & rhs);
    CLdapFilter & And(const wstring & rhs);
    CLdapFilter & And(const wchar_t * rhs);

    // Logical OR
    CLdapFilter & Or(const CLdapFilter & rhs);
    CLdapFilter & Or(const wstring & rhs);
    CLdapFilter & Or(const wchar_t * rhs);

    // Logical NOT
    CLdapFilter & Not();

    // Filter presence
    bool HasFilter() const;

    // Get functions
    const wstring & GetString();
    const wchar_t * GetCString();

    // Clear
    void Clear();

#ifndef LDAPFILTER_SUPPORT_EXCEPTION
    const wchar_t * GetLastErrorString() const;
    size_t GetLastError() const;
    size_t GetLastErrorPos() const;
    void ClearLastError();
#endif

protected:
    // Parser functions
    int LdapEscapable(wchar_t wc1, wchar_t wc2 = L'\0');
    size_t ValidateAttributeType(wstring & wsAttr, size_t sAbsIniPos);
    size_t ValidateAttributeValue(const wstring & wsValue, 
				  size_t sAbsIniPos);
    size_t ParseCondition(wstring &wsFilterCond, size_t sAbsIniPos);
    size_t FullParser(const wstring &wsFilter, 
		      CLdapFilter::CFilterTree & FilterTree, 
		      size_t sAbsIniPos);
    size_t ParseString(const wstring & wsFilter);

    // SetLastError
    size_t SetLastError(size_t sPos, size_t sErrorCode);

    // GenerateString
    void GenerateString(const CFilterTree & Filter, 
			wstring & wsFilter);

private:
    CFilterTree m_FilterTree;
    
    wstring m_wsGeneratedFilter;
    bool m_fModified;

#ifndef LDAPFILTER_SUPPORT_EXCEPTION
    // Only used if we are NOT throwing exceptions
    size_t m_sLastError;
    size_t m_sErrorPosition;
#endif

};

//
// Operator &&
//

// CLdapFilter
CLdapFilter operator&&(const CLdapFilter lhs, 
		       const CLdapFilter & rhs);

// wstring
inline CLdapFilter operator&&(const CLdapFilter lhs, 
			      const wstring & rhs)
{    return operator&&(lhs,CLdapFilter(rhs)); }

inline CLdapFilter operator&&(const wstring & lhs, 
			      const CLdapFilter & rhs)
{    return operator&&(CLdapFilter(lhs), rhs); }

// wchar_t *
inline CLdapFilter operator&&(const CLdapFilter lhs, 
			      const wchar_t * rhs)
{    return operator&&(lhs,CLdapFilter(rhs)); }

inline CLdapFilter operator&&(const wchar_t * lhs, 
			      const CLdapFilter & rhs)
{    return operator&&(CLdapFilter(lhs), rhs); }


//
// Operator ||
//

// CLdapFilter
CLdapFilter operator||(const CLdapFilter lhs, 
		       const CLdapFilter & rhs);

// wstring
inline CLdapFilter operator||(const CLdapFilter lhs, 
			      const wstring & rhs)
{    return operator||(lhs,CLdapFilter(rhs)); }

inline CLdapFilter operator||(const wstring & lhs, 
			      const CLdapFilter & rhs)
{    return operator||(CLdapFilter(lhs), rhs); }

// wchar_t *
inline CLdapFilter operator||(const CLdapFilter lhs, 
			      const wchar_t * rhs)
{    return operator||(lhs,CLdapFilter(rhs)); }

inline CLdapFilter operator||(const wchar_t * lhs, 
			      const CLdapFilter & rhs) 
{    return operator||(CLdapFilter(lhs), rhs); }

//
// Operator !
//
CLdapFilter operator!(const CLdapFilter &rhs);


#endif