/******************************************************************************

  IR unit interface

  This is a thin layer between the serial port and the knowledge of IR key
  presses. It performs no buffering other than that provided by the underlying
  COM port, and merely queues up sequences of 6 characters before translating
  them to an index into a vector.

  Note that the COM port is not opened until the first read, so that potential
  hangs occur in the separate thread.

                      Copyright 1999 Gavin Smyth
                 http://www.beesknees.freeserve.co.uk

******************************************************************************/

#include "common.h"
#include "irman.h"
#include "registry.h"

// Registry keys
const static TCHAR RegPort[] = _T("Port");
const static TCHAR RegDelay[] = _T("InterKeyDelay");

Irman::Irman( const TCHAR* regKey_,  int numKeys_ ) :
        regKey( regKey_ ), numKeys( numKeys_ ),
        keyCodes( new KeyCode[ numKeys_ ] ),
        comPort( INVALID_HANDLE_VALUE ),
        keyTime( GetTickCount() ), setKey( -1 ),
        ioCompletion( CreateEvent( NULL, TRUE, FALSE, NULL ) )
{
  // Read the port name, inter key delay and all the key codes from
  // the registry, and then fire up the device

  RegistryKey reg( HKEY_LOCAL_MACHINE, regKey );
  reg.Read( RegPort, comPortName, sizeof( comPortName ) / sizeof( TCHAR ), _T("COM2") );
  interKeyDelay = reg.Read( RegDelay, 500 );
  for( int i = 0; i < numKeys; ++i )
  {
    static KeyCode blankKeyCode;
    TCHAR num[ 16 ];
    wsprintf( num, _T("%03d"), i );
    reg.Read( num, keyCodes[ i ].code, sizeof( KeyCode ), &blankKeyCode );
#   ifdef VERBOSE
      TCHAR buff[ 40 ];
      wsprintf( buff, "Key %d is %02X %02X %02X %02X %02X %02X\n", i,
                      keyCodes[ i ].code[0], keyCodes[ i ].code[1], keyCodes[ i ].code[2],
                      keyCodes[ i ].code[3], keyCodes[ i ].code[4], keyCodes[ i ].code[5] );
      OutputDebugString( buff );
#   endif
  }

  // Don't open the port yet in case it hangs...
}

Irman::~Irman()
{
  // Close everything down, and write back to the registry

  Close();
  if( ioCompletion != INVALID_HANDLE_VALUE )
  {
    CloseHandle( ioCompletion );
    ioCompletion = INVALID_HANDLE_VALUE;
  }

  RegistryKey reg( HKEY_LOCAL_MACHINE, regKey );
  reg.Write( RegPort, comPortName );
  reg.Write( RegDelay, interKeyDelay );
  for( int i = 0; i < numKeys; ++i )
  {
    TCHAR num[ 16 ];
    wsprintf( num, _T("%03d"), i );
    reg.Write( num, keyCodes[ i ].code, sizeof( KeyCode ) );
  }

  delete [] keyCodes;
}

bool Irman::Open()
{
  const TCHAR* error = _T("Error opening device");
  comPort = CreateFile( comPortName, GENERIC_READ | GENERIC_WRITE,
                        0, NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL );
  if( comPort == INVALID_HANDLE_VALUE )
    MessageBox( NULL, _T("Could not open COM port - maybe something else is using it"), error, MB_ICONEXCLAMATION | MB_OK );
  else
  {
    DCB dcb;
    if( Verify( GetCommState( comPort, &dcb ) ) )
    {
      dcb.BaudRate    = CBR_9600;
      dcb.fParity     = 0;
      dcb.Parity      = NOPARITY;
      dcb.ByteSize    = 8;
      dcb.StopBits    = ONESTOPBIT;
      dcb.fDtrControl = DTR_CONTROL_DISABLE;
      dcb.fRtsControl = RTS_CONTROL_DISABLE;
      if( Verify( SetCommState( comPort, &dcb ) ) )
      {
        PowerOff();           // Just in case it was already on
        Sleep( 200 );
        PowerOn();
        Sleep( 100 );         // Time for the output to settle
        Flush();              // Remove power up garbage
        WriteWait( "I", 1 );  // These strings must be ASCII, not Unicode
        Sleep( 2 );           // Need to have >500us between the 'I' & the 'R'
        WriteWait( "R", 1 );

        char data[ 2 ];
        if( ReadWait( data, 2 ) && data[ 0 ] == 'O' && data[ 1 ] == 'K' )
          return true;
        else
          MessageBox( NULL, _T("Irman not responding"), error, MB_ICONEXCLAMATION | MB_OK );
      }
    }
  }

  // To get this far, something must have gone wrong
  Close();
  return false;
}

void Irman::Close()
{
  if( comPort != INVALID_HANDLE_VALUE )
  {
    Verify( CloseHandle( comPort ) );
    comPort = INVALID_HANDLE_VALUE;
  }
}

void Irman::Port( const TCHAR* comPort )
{
  _tcsnccpy( comPortName, comPort, sizeof( comPortName ) / sizeof( TCHAR ) );
  comPortName[ sizeof( comPortName ) / sizeof( TCHAR ) - 1 ] = 0;

  // Reopen the port if the name changed - I could have checked the new
  // and old names and if they were the same, skip the reopen. However,
  // this way, I can force a recover from a "stuck" I/O port...

  Close();
}

void Irman::SetKey( int key )
{
  if( comPort == INVALID_HANDLE_VALUE )
  {
    MessageBox( NULL, _T("COM port not valid - can't configure"), _T("IR Configuration Error"), MB_ICONEXCLAMATION | MB_OK );
    return;
  }

  if( key < 0 || key > numKeys )
  {
    return;
  }

  // Just indicate to the reading function that it should store the next
  // sequence instead of matching it

  setKey = key;
}

int Irman::Key()
{
  if( comPort == INVALID_HANDLE_VALUE && !Open() )
    return -1;

  // The Irman reports a number of sequences for each key - chuck away
  // old ones before reading the next key.
  Flush();

  KeyCode key;

  // Loop for a minimum time, to get rid of old duplicate/inverted messages
  unsigned long startTime = keyTime;
  do
  {
    if( !Read( key ) )
      return -1;
  } while( keyTime - startTime < interKeyDelay );

  // Now, key contains a valid code sequence, so do something with it

  if( setKey != -1 )
  {
    // If we're in record mode, just use this sequence for the relevant
    // entry in keyCodes[]
    keyCodes[ setKey ] = key;
    int retVal = setKey;
    setKey = -1;
    return retVal;
  }
  else
  {
    // If we're not in record mode, scan the list to find a match,
    // and repeat for up to the inter key period before giving up and
    // admitting it's unrecognised - the reason for the loop is to
    // catch any inversions along the way
    startTime = keyTime;
    do
    {
      for( int i = 0; i < numKeys; ++i )
        if( key == keyCodes[ i ] )
          return i;
      if( !Read( key ) )
        return -1;
    } while( GetTickCount() - startTime < interKeyDelay );
    return -1; // No key found
  }
}

bool Irman::Read( KeyCode& key )
{
  bool success = ReadWait( key.code, sizeof( key.code ) );
# ifdef VERBOSE
    if( success )
    {
      TCHAR buff[ 40 ];
      wsprintf( buff, "Code %02X %02X %02X %02X %02X %02X\n",
                      key.code[0], key.code[1], key.code[2],
                      key.code[3], key.code[4], key.code[5] );
      OutputDebugString( buff );
    }
# endif
  keyTime = GetTickCount();
  return success;
}

void Irman::Interrupt()
{
  // On NT or 98, this should use CancelIo, but that function isn't present
  // on 95, so bodge it by signalling the event...
  // CancelIo( comPort );
  SetEvent( ioCompletion );
}

// It's a bit of a shame that there is some much code repetition in reading
// and writing. I did experiment with a common function into which I passed
// a pointer to either ReadFile or WriteFile, but C++'s type system confounded
// that by making it impossible to write legal code that would cope with the
// const arguments of WriteFile vs the non-const ones of ReadFile.

bool Irman::ReadWait( void* data, unsigned long size )
{
  if( comPort == INVALID_HANDLE_VALUE )
    return false;

  OVERLAPPED ov;
  ZeroMemory( &ov, sizeof( ov ) );
  ov.hEvent = ioCompletion;

  // If ReadFile doesn't succeed, but the error is ERROR_IO_PENDING, we
  // just wait for the operation to complete.

  DWORD length;
  if( !ReadFile( comPort, data, size, &length, &ov ) )
  {
    if( GetLastError() == ERROR_IO_PENDING )
    {
      if( Verify( GetOverlappedResult( comPort, &ov, &length, TRUE ) ) == FALSE )
        return false;
    }
    else
    {
      Verify( FALSE ); // Read failed
      return false;
    }
  }

  // At this point, the buffer contains the expected data
  return size == length;
}

// Very similar to ReadWait...

bool Irman::WriteWait( const void* data, unsigned long size )
{
  if( comPort == INVALID_HANDLE_VALUE )
    return false;

  OVERLAPPED ov;
  ZeroMemory( &ov, sizeof( ov ) );
  ov.hEvent = ioCompletion;

  DWORD length;
  if( !WriteFile( comPort, data, size, &length, &ov ) )
  {
    if( GetLastError() == ERROR_IO_PENDING )
    {
      if( Verify( GetOverlappedResult( comPort, &ov, &length, TRUE ) ) == FALSE )
        return false;
    }
    else
    {
      Verify( FALSE ); // Write failed
      return false;
    }
  }

  return size == length;
}

// Power the Irman device by twiddling the control lines

void Irman::PowerOff() const
{
  Verify( EscapeCommFunction( comPort, CLRDTR ) );
  Verify( EscapeCommFunction( comPort, CLRRTS ) );
}

void Irman::PowerOn() const
{
  Verify( EscapeCommFunction( comPort, SETDTR ) );
  Verify( EscapeCommFunction( comPort, SETRTS ) );
}

// Throw away anything in the COM port buffers, and set the ignore time
// to start from now.

void Irman::Flush()
{
  Verify( PurgeComm( comPort, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR ) );
  keyTime = GetTickCount();
}
