/******************************************************************************

  IR unit interface

  This is a thin layer between the serial port and the knowledge of IR key
  presses. It performs no buffering other than that provided by the underlying
  COM port, and merely queues up sequences of 6 characters before translating
  them to an index into a vector.

  It reads the key code sequences from the registry, and takes a list of
  key names from the calling program. If no key mapping has been defined
  (or even it it has), the SetKey() method can be used to create this
  mapping.

                      Copyright 1999 Gavin Smyth
                 http://www.beesknees.freeserve.co.uk

******************************************************************************/

#ifndef _h_irman_h_
#define _h_irman_h_

class Irman
{
public:
  // Initialise from registry (or take defaults)
  Irman( const TCHAR* regKey,       // Registry key under which to find the values
         int numKeys );             // Number of keys on the Remote

  // Shutdown, and rewrite configuration to the registry
  ~Irman();

  // Get and set the port name (COMx)
  const TCHAR* Port() const { return comPortName; }
  void Port( const TCHAR* comPort );

  // Get and set the inter-key delay (milliseconds)
  unsigned long Delay() const { return interKeyDelay; }
  void Delay( unsigned long d ) { interKeyDelay = d; }

  // Wait for a data packet to be received and return the index into
  // the vector that represents it (or -1 if not recognised)
  int Key();

  // Trigger the next received key sequence to be stored for the
  // indicated key code
  void SetKey( int key );

  // Interrupt the current read - used from a separate task
  void Interrupt();

private:
  // Each Irman key is decoded to a sequence of 6 bytes
  struct KeyCode
  {
    unsigned char code[ 6 ];

    bool operator==( const KeyCode& key )
    {
      for( int i = 0; i < sizeof( code ); ++i )
        if( key.code[ i ] != code[ i ] )
          return false;
      return true;
    }
  };

  TCHAR comPortName[ 5 ];  // contains COMx\0
  volatile HANDLE comPort; // Handle to the opened COM port
  HANDLE ioCompletion;     // Handle used for overlapped I/O

  // Where to read/write values - passed to constructor
  const TCHAR* regKey;

  // How many keys on the remote - passed to constructor
  int numKeys;

  // Codes corresponding to each key to be recognised (numKeys long)
  KeyCode* keyCodes;

  // Open and close the COM port
  bool Open();
  void Close();

  // Read a complete Irman sequence
  bool Read( KeyCode& );

  // Time (ms) last Irman sequence read
  unsigned long keyTime;

  // Key code to which to set next read key
  volatile int setKey;

  // Low level (blocking, but interruptable via Interrupt() above)
  // COM port read and write
  bool ReadWait( void* data, unsigned long size );
  bool WriteWait( const void* data, unsigned long size );

  // Waggle the control lines to power up or down the Irman
  void PowerOn() const;
  void PowerOff() const;

  // Discard any characters in the COM port buffers
  void Flush();

  // Time to wait from reading one key ro the next
  volatile unsigned long interKeyDelay;

  // Disable copying
  Irman( const Irman& );
  Irman& operator=( const Irman& );  
};

#endif // _h_irman_h_
