/******************************************************************************

  Irman Winamp controller main source

  If this is built with PLUGIN defined, it will produce a Winamp plugin,
  otherwise it produces a standalone application.

                      Copyright 1999 Gavin Smyth
                 http://www.beesknees.freeserve.co.uk

******************************************************************************/

#include "common.h"
#include "winampmgr.h"
#include "irman.h"
#include "resource.h"
#include <process.h>

#define UID 8877                   // Fairly arbitrary id for the tray icon
#define NOT_MSG ( WM_USER + 33 )   // Tray message id (arbitrary value)

static HMENU menu;                 // Popup menu

// There are two icons in the system tray, and iconNumber indicates
// which to display next

static HICON icon[ 2 ];
static int iconNumber;

// Names of the remote keys recognised - should be in the same order as 
// the configure dialog box buttons, and the same order as the Irman key
// codes.

static const TCHAR* keyStrings[] =
{
  _T("Next Track"),
  _T("Previous Track"),
  _T("Increase Volume"),
  _T("Decrease Volume"),
  _T("Shutdown PC")
};
static const int numKeys = sizeof( keyStrings ) / sizeof( keyStrings[ 0 ] );

// Singleton Winamp controller and Irman interface

WinampManager mgr;
Irman ir( _T("Software\\BeesKnees\\Irman"), numKeys );

volatile HWND mainWnd;       // Handle to the (invisible) application window
volatile HWND configureWnd;  // Handle to the configuration dialog box: if
                             // this is non-null, the dialog is visible
volatile bool recording;     // True if we're waiting for a key to be recorded
volatile bool testing;       // True if we're testing instead of acting on keys

volatile bool going = true;  // The reader loop runs while this is set

static void Reader( void* )
{
  while( going )
  {
    // Rather sloppily, this does not bother to guard against changes in the
    // COM port (eg, a different one selected by the code below), but it seems
    // to work OK...
    int key = ir.Key();

    // Toggle the tray icon - only visible if it's not a plugin
#   ifndef PLUGIN
      NOTIFYICONDATA not;
      not.cbSize = sizeof( not );
      not.hWnd = mainWnd;
      not.uID = UID;
      not.uFlags = NIF_ICON;
      not.hIcon = icon[ iconNumber = !iconNumber ];
      Shell_NotifyIcon( NIM_MODIFY, &not );
#   endif

    if( configureWnd )    // ie, if we're running the configure dialog
    {
      if( recording )     // ie, we need to respond to a completed record
      {
        SetDlgItemText( configureWnd, IDC_STATUS, key != -1 ? _T("OK") : _T("Failed") );
        recording = false;
      }
      else if( testing )  // ie, echo, don't act
      {
        TCHAR buff[ 64 ];
        wsprintf( buff, _T("Read %d (%s)"), key, key != -1 ? keyStrings[ key ] : _T("Undefined") );
        SetDlgItemText( configureWnd, IDC_STATUS, buff );
      }
    }
    else
    {
#     ifdef VERBOSE
        TCHAR buff[ 64 ];
        wsprintf( buff, _T("Read %d (%s)\n"), key, key != -1 ? keyStrings[ key ] : _T("Undefined") );
        OutputDebugString( buff );
#     endif

      // Just perform the appropriate action
      switch( key )
      {
        case 0:
          mgr.Skip( true );
          break;
        case 1:
          mgr.Skip( false );
          break;
        case 2:
          mgr.Volume( true );
          break;
        case 3:
          mgr.Volume( false );
          break;
        case 4:
          mgr.Exit();
          break;
      }
    }
  } // end while going
}

BOOL CALLBACK ConfigProc( HWND hwnd, UINT msg, WPARAM wParam, LPARAM /*lParam*/ )
{
  switch( msg )
  {
    case WM_INITDIALOG:
      {
        configureWnd = hwnd;    // Setting this indicates that configuration is active
        const TCHAR* port = ir.Port();
        if( _tcslen( port ) == 4 && port[ 3 ] >= _T('1') && port[ 3 ] <= _T('4') )
          CheckRadioButton( hwnd, IDC_COM1, IDC_COM4, IDC_COM1 + port[ 3 ] - _T('1') );
        SetDlgItemInt( hwnd, IDC_DELAY, ir.Delay(), FALSE );
      }
      return TRUE;

    case WM_COMMAND:
      switch( LOWORD( wParam ) )
      {
        case IDOK:
          {
            BOOL success;
            UINT d = GetDlgItemInt( hwnd, IDC_DELAY, &success, FALSE );
            if( success )
              ir.Delay( d );
            EndDialog( hwnd, 0 );
            configureWnd = NULL;
          }
          return TRUE;

        case IDC_COM1:
        case IDC_COM2:
        case IDC_COM3:
        case IDC_COM4:
          {
            // I could check that the radio button hasn't already been
            // selected, but not doing this has the side-effect of
            // reinitialising the port if necessary...
            TCHAR buff[ 8 ];
            wsprintf( buff, _T("COM%d"), LOWORD( wParam ) - IDC_COM1 + 1 );
            ir.Port( buff );
          }
          return TRUE;

        case IDC_NEXT_TRACK:
        case IDC_PREV_TRACK:
        case IDC_VOL_UP:
        case IDC_VOL_DOWN:
        case IDC_SHUTDOWN:
          {
            int code = LOWORD( wParam ) - IDC_NEXT_TRACK;
            TCHAR buff[ 64 ];
            wsprintf( buff, _T("Press remote key for %s"), keyStrings[ code ] );
            SetDlgItemText( hwnd, IDC_STATUS, buff );
            recording = true;         // Initiate a key record
            ir.SetKey( code );
          }
          return TRUE;

        case IDC_CHECK_TEST:
          testing = IsDlgButtonChecked( hwnd, IDC_CHECK_TEST ) == BST_CHECKED;
          return TRUE;
      }
      break;
  }
  return FALSE;
}

/*****************************************************************************

  Main (invisible) window procedure. Just has to respond to create (spawn off
  sub process), destroy and tray commands (toggle visibility).

  This is only defined if we're not building a Winamp plugin.

*****************************************************************************/

#ifndef PLUGIN

LRESULT CALLBACK WndProc( HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam )
{

  switch( msg )
  {
    case WM_CREATE:
      _beginthread( Reader, 0, NULL );
      break;

    case WM_DESTROY:
      going = false;  // This will kill off the reader
      ir.Interrupt();
      PostQuitMessage( 0 );
      return 0;

    case WM_COMMAND:
      switch( LOWORD( wParam ) )
      {
        case ID_CONFIGURE:
          DialogBox( reinterpret_cast< HINSTANCE >( GetWindowLong( hwnd, GWL_HINSTANCE ) ),
                     MAKEINTRESOURCE( IDD_CONFIGURE ), hwnd, ConfigProc );
          break;

        case ID_QUIT:
          SendMessage( hwnd, WM_CLOSE, 0, 0 );
          return 0;
      }
      break;

    case NOT_MSG:
      // The only message I expect from the system tray is a mouse click,
      // in which case I pop up a menu at that cursor position
      switch( lParam )
      {
        case WM_LBUTTONDOWN:
        case WM_RBUTTONDOWN:
          {
            POINT p;
            GetCursorPos( &p );
            SetForegroundWindow( hwnd );
            TrackPopupMenu( menu, TPM_RIGHTALIGN | TPM_RIGHTBUTTON, p.x, p.y, 0, hwnd, NULL) ;
          }
          break;
      }
      return 0;
  }
  return DefWindowProc( hwnd, msg, wParam, lParam );
}

int WINAPI _tWinMain( HINSTANCE hInstance, HINSTANCE /*prev*/, LPSTR /*cmdline*/, int /*cmdShow*/ )
{
  icon[ 0 ] = LoadIcon( hInstance, MAKEINTRESOURCE( IDI_ICON ) );
  icon[ 1 ] = LoadIcon( hInstance, MAKEINTRESOURCE( IDI_ICON_INV ) );

  WNDCLASS wc;
  wc.style         = 0;
  wc.lpfnWndProc   = WndProc;
  wc.cbClsExtra    = 0;
  wc.cbWndExtra    = 0;
  wc.hInstance     = hInstance;
  wc.hIcon         = icon[ 0 ];
  wc.hCursor       = LoadCursor( NULL, IDC_ARROW );
  wc.hbrBackground = (HBRUSH)( COLOR_WINDOW + 1 );
  wc.lpszMenuName  = NULL;
  wc.lpszClassName = _T("WinampIR");

  if( !RegisterClass( &wc ) )
    return 0;

  mainWnd = CreateWindow( wc.lpszClassName, wc.lpszClassName,
                          WS_POPUP,
                          CW_USEDEFAULT, 0,
                          CW_USEDEFAULT, 0,
                          NULL, NULL,
                          hInstance, NULL );
  if( !mainWnd )
    return 0;

  menu = LoadMenu( hInstance, MAKEINTRESOURCE( IDR_MENU ) );
  menu = GetSubMenu( menu, 0 );

  // Populate the system tray
  NOTIFYICONDATA not;
  not.cbSize = sizeof( not );
  not.hWnd = mainWnd;
  not.uID = UID;
  not.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
  not.uCallbackMessage = NOT_MSG;
  not.hIcon = wc.hIcon;
  _tcsncpy( not.szTip, _T("Winamp IR controller"), 20 );
  Shell_NotifyIcon( NIM_ADD, &not );

  MSG msg;
  while( GetMessage( &msg, NULL, 0, 0 ) )
  {
    TranslateMessage( &msg );
    DispatchMessage( &msg );
  }

  // Release all Windows resources grabbed at startup
  Shell_NotifyIcon( NIM_DELETE, &not );
  DestroyIcon( icon[ 0 ] );
  DestroyIcon( icon[ 1 ] );

  return msg.wParam;
}

#else // PLUGIN defined

/******************************************************************************

  The following section implements the Winamp interface

******************************************************************************/

#include "gen.h"

static int init();
static void config();
static void quit();

static winampGeneralPurposePlugin plugin =
{
  GPPHDR_VER,
  "Irman control interface 1.0 (gen_bees.dll)",
  init,
  config,
  quit,
  NULL,
  NULL
};

static int init()
{
  mgr.SetWinamp( plugin.hwndParent );
  _beginthread( Reader, 0, NULL );
  return 0;
}

static void config()
{
  DialogBox( plugin.hDllInstance, MAKEINTRESOURCE( IDD_CONFIGURE ),
             plugin.hwndParent, ConfigProc );
}

static void quit()
{
  going = false;
  ir.Interrupt();
  Sleep( 100 );  // Give the thread time to die quietly
}

extern "C" __declspec( dllexport ) winampGeneralPurposePlugin* winampGetGeneralPurposePlugin()
{
  return &plugin;
}

#endif // PLUGIN
